@@ -64,7 +64,7 @@ struct ReturnTypeAdapter<void, Args...>
 template<typename R, typename... Args>
 struct CallFunctor
 {
-  using return_type = typename std::remove_const<decltype(ReturnTypeAdapter<R, Args...>()(std::declval<const void*>(), std::declval<static_julia_type<Args>>()...))>::type;
+  using return_type = std::remove_const_t<decltype(ReturnTypeAdapter<R, Args...>()(std::declval<const void*>(), std::declval<static_julia_type<Args>>()...))>;
 
   static return_type apply(const void* functor, static_julia_type<Args>... args)
   {
@@ -93,7 +93,7 @@ struct NeedConvertHelper
 {
   bool operator()()
   {
-    for(const bool b : {std::is_same<static_julia_type<Args>,Args>::value...})
+    for(const bool b : {std::is_same_v<static_julia_type<Args>,Args>...})
     {
       if(!b)
         return true;
@@ -119,7 +119,7 @@ BoxedValue<T> create(ArgsT&&... args)
 {
   jl_datatype_t* dt = julia_type<T>();
   assert(jl_is_mutable_datatype(dt));
-  
+
 
   T* cpp_obj = new T(std::forward<ArgsT>(args)...);
 
@@ -391,7 +391,7 @@ struct ParameterList
     {
       if(paramlist[i] == nullptr)
       {
-        std::vector<std::string> typenames({(typeid(ParametersT).name())...}); 
+        std::vector<std::string> typenames({(typeid(ParametersT).name())...});
         throw std::runtime_error("Attempt to use unmapped type " + typenames[i] + " in parameter list");
       }
     }
@@ -424,69 +424,45 @@ namespace detail
     }
   };
 
-  template<typename... T>
-  constexpr bool has_type = false;
-
-  template<typename T1, typename T2, typename... ParametersT>
-  constexpr bool has_type<T1, ParameterList<T2, ParametersT...>> = has_type<T1, ParameterList<ParametersT...>>;
+  template<typename T, typename... ParametersT>
+  struct has_type : std::bool_constant<(... || std::is_same_v<T, ParametersT>)> {};
 
   template<typename T, typename... ParametersT>
-  constexpr bool has_type<T, ParameterList<T, ParametersT...>> = true;
+  constexpr bool has_type_v = has_type<T, ParametersT...>::value;
 
   template<typename T1, typename T2>
-  constexpr bool has_type<T1, T2, ParameterList<>> = false;
-
-  template<bool, typename T1, typename T2>
-  struct ConditionalAppend
-  {
-  };
-
-  template<typename T, typename... ParametersT>
-  struct ConditionalAppend<true, T, ParameterList<ParametersT...>>
-  {
-    using type = ParameterList<ParametersT..., T>;
-  };
-
-  template<typename T, typename... ParametersT>
-  struct ConditionalAppend<false, T, ParameterList<ParametersT...>>
-  {
-    using type = ParameterList<ParametersT...>;
-  };
-
-  template<typename... T>
-  struct RemoveDuplicates
-  {
-    using type = ParameterList<>;
-  };
-
-  template<typename T1, typename... ParametersT>
-  struct RemoveDuplicates<ParameterList<T1,ParametersT...>>
-  {
-    using type = typename RemoveDuplicates<ParameterList<T1>, ParameterList<ParametersT...>>::type;
-  };
-
-  template<typename ResultT, typename T1, typename... ParametersT>
-  struct RemoveDuplicates<ResultT, ParameterList<T1,ParametersT...>>
-  {
-    using type = typename RemoveDuplicates<typename ConditionalAppend<!has_type<T1,ResultT>,T1,ResultT>::type, ParameterList<ParametersT...>>::type;
-  };
-
-  template<typename ResultT>
-  struct RemoveDuplicates<ResultT, ParameterList<>>
-  {
-    using type = ResultT;
-  };
-
-  template<typename T1, typename T2>
-  struct CombineParameterLists
-  {
-  };
+  struct CombineParameterLists;
 
   template<typename... Params1, typename... Params2>
   struct CombineParameterLists<ParameterList<Params1...>, ParameterList<Params2...>>
   {
     using type = ParameterList<Params1..., Params2...>;
   };
+
+  template<typename... T>
+  struct RemoveDuplicates;
+
+  // empty/bottom case
+  template<>
+  struct RemoveDuplicates<ParameterList<>>
+  {
+    using type = ParameterList<>;
+  };
+
+  template<typename T, typename... Rest>
+  struct RemoveDuplicates<ParameterList<T, Rest...>>
+  {
+  private:
+    // recursive splitting of remaining parameters
+    using rest_t = typename RemoveDuplicates<ParameterList<Rest...>>::type;
+
+  public:
+    using type = std::conditional_t<
+      has_type_v<T, Rest...>,
+      rest_t,  // T appears later, skip it
+      typename CombineParameterLists<rest_t, ParameterList<T>>::type  // T is unique so far, append it
+    >;
+  };
 }
 
 template<typename ParametersT, typename F>
@@ -543,13 +519,13 @@ template<> inline std::string fixed_int_type_name<uint64_t>() { return "uint64_t
 
 /// Trait to allow user-controlled disabling of the default constructor
 template <typename T>
-struct DefaultConstructible : std::bool_constant<std::is_default_constructible<T>::value && !std::is_abstract<T>::value>
+struct DefaultConstructible : std::bool_constant<std::is_default_constructible_v<T> && !std::is_abstract_v<T>>
 {
 };
 
 /// Trait to allow user-controlled disabling of the copy constructor
 template <typename T>
-struct CopyConstructible : std::bool_constant<std::is_copy_constructible<T>::value && !std::is_abstract<T>::value>
+struct CopyConstructible : std::bool_constant<std::is_copy_constructible_v<T> && !std::is_abstract_v<T>>
 {
 };
 
@@ -607,7 +583,7 @@ public:
 
   /// Define a new function. Overload for lambda
   template<typename LambdaT, typename... Extra,
-           std::enable_if_t<detail::has_call_operator<LambdaT>::value && !std::is_member_function_pointer<LambdaT>::value, bool> = true>
+           std::enable_if_t<detail::has_call_operator<LambdaT>::value && !std::is_member_function_pointer_v<LambdaT>, bool> = true>
   FunctionWrapperBase& method(const std::string& name, LambdaT&& lambda, Extra... extra)
   {
     detail::ExtraFunctionData extraData = detail::parse_attributes(extra...);
@@ -630,7 +606,7 @@ public:
   template<typename T, typename R, typename LambdaT, typename... ArgsT, typename... Extra>
   void constructor(jl_datatype_t* dt, LambdaT&& lambda, R(LambdaT::*)(ArgsT...) const, Extra... extra)
   {
-    static_assert(std::is_same<T*,R>::value, "Constructor lambda function must return a pointer to the constructed object, of the correct type");
+    static_assert(std::is_same_v<T*,R>, "Constructor lambda function must return a pointer to the constructed object, of the correct type");
     detail::ExtraFunctionData extraData = detail::parse_attributes<false,true>(extra...);
     FunctionWrapperBase &new_wrapper = add_lambda("dummy", [=](ArgsT... args)
     {
@@ -832,6 +808,7 @@ private:
   std::vector<jl_datatype_t*> m_box_types;
 
   template<class T> friend class TypeWrapper;
+  template<typename T, typename... AppliedTypesT> friend class ParametricTypeWrappers;
 };
 
 template<typename T>
@@ -870,14 +847,22 @@ struct BuildParameterList<T<ParametersT...>>
 namespace detail
 {
   template<typename... Types>
-  struct DoApply;
-
-  template<>
-  struct DoApply<>
+  struct DoApply
   {
     template<typename WrapperT, typename FunctorT>
-    void operator()(WrapperT&, FunctorT&&)
+    void operator()(WrapperT& w, FunctorT&& ftor)
     {
+      (..., DoApply<Types>()(w, std::forward<FunctorT>(ftor)));
+    }
+  };
+
+  template<typename... Types>
+  struct DoApply<ParameterList<Types...>>
+  {
+    template<typename WrapperT, typename FunctorT>
+    void operator()(WrapperT& w, FunctorT&& ftor)
+    {
+      (..., DoApply<Types>()(w, std::forward<FunctorT>(ftor)));
     }
   };
 
@@ -890,39 +875,26 @@ namespace detail
       w.template apply<AppT>(std::forward<FunctorT>(ftor));
     }
   };
-
-  template<typename... Types>
-  struct DoApply<ParameterList<Types...>>
-  {
-    template<typename WrapperT, typename FunctorT>
-    void operator()(WrapperT& w, FunctorT&& ftor)
-    {
-      DoApply<Types...>()(w, std::forward<FunctorT>(ftor));
-    }
-  };
-
-  template<typename T1, typename... Types>
-  struct DoApply<T1, Types...>
-  {
-    template<typename WrapperT, typename FunctorT>
-    void operator()(WrapperT& w, FunctorT&& ftor)
-    {
-      DoApply<T1>()(w, std::forward<FunctorT>(ftor));
-      DoApply<Types...>()(w, std::forward<FunctorT>(ftor));
-    }
-  };
 }
 
 /// Execute a functor on each type
 template<typename... Types>
-struct ForEachType;
-
-template<>
-struct ForEachType<>
+struct ForEachType
 {
   template<typename FunctorT>
-  void operator()(FunctorT&&)
+  void operator()(FunctorT&& ftor)
   {
+    (..., ForEachType<Types>()(std::forward<FunctorT>(ftor)));
+  }
+};
+
+template<typename... Types>
+struct ForEachType<ParameterList<Types...>>
+{
+  template<typename FunctorT>
+  void operator()(FunctorT&& ftor)
+  {
+    (..., ForEachType<Types>()(std::forward<FunctorT>(ftor)));
   }
 };
 
@@ -934,81 +906,116 @@ struct ForEachType<AppT>
   {
 #ifdef _MSC_VER
     ftor.operator()<AppT>();
-#else 
+#else
     ftor.template operator()<AppT>();
 #endif
   }
 };
 
-template<typename... Types>
-struct ForEachType<ParameterList<Types...>>
-{
-  template<typename FunctorT>
-  void operator()(FunctorT&& ftor)
-  {
-    ForEachType<Types...>()(std::forward<FunctorT>(ftor));
-  }
-};
-
-template<typename T1, typename... Types>
-struct ForEachType<T1, Types...>
-{
-  template<typename FunctorT>
-  void operator()(FunctorT&& ftor)
-  {
-    ForEachType<T1>()(std::forward<FunctorT>(ftor));
-    ForEachType<Types...>()(std::forward<FunctorT>(ftor));
-  }
-};
-
 template<typename T, typename FunctorT>
 void for_each_type(FunctorT&& f)
 {
   ForEachType<T>()(f);
 }
 
+/// Type-level container for accumulating parameter packs during CombineTypes recursion.
+/// Semantically equivalent to ParameterList but with a distinct typename to enable
+/// template specialization in CombineTypes.
 template<typename... Types>
 struct UnpackedTypeList
 {
 };
 
+template<typename... Types>
+struct FlattenParameterLists;
+
+// Base case: single non-ParameterList type
+template<typename T>
+struct FlattenParameterLists<T>
+{
+  using flatten = ParameterList<T>;
+};
+
+// Single ParameterList: flatten it recursively
+template<typename... Types>
+struct FlattenParameterLists<ParameterList<Types...>>
+{
+  using flatten = typename FlattenParameterLists<Types...>::flatten;
+};
+
+// Multiple types: recursively flatten each and concatenate
+template<typename T1, typename T2, typename... Rest>
+struct FlattenParameterLists<T1, T2, Rest...>
+{
+  using flatten = typename detail::CombineParameterLists<
+      typename FlattenParameterLists<T1>::flatten,
+      typename FlattenParameterLists<T2, Rest...>::flatten
+    >::type;
+};
+
+/// Compute the cartesian product of multiple ParameterLists, applying ApplyT to each combination.
+///
+/// Given ApplyT and N ParameterLists, produces a ParameterList containing ApplyT applied to all
+/// possible combinations of one type from each input list. ApplyT must provide a template type
+/// alias `apply` that accepts the unpacked combination.
+///
+/// Example from parametric.cpp:
+///   CombineTypes<ApplyType<Foo3>, ParameterList<int32_t, double>, ParameterList<P1,P2,bool>, ParameterList<float>>::type
+///   => ParameterList<Foo3<int32_t,P1,float>, Foo3<int32_t,P2,float>, Foo3<int32_t,bool,float>,
+///                    Foo3<double,P1,float>, Foo3<double,P2,float>, Foo3<double,bool,float>>
 template<typename ApplyT, typename... TypeLists>
 struct CombineTypes;
 
-template<typename ApplyT, typename... UnpackedTypes>
-struct CombineTypes<ApplyT, UnpackedTypeList<UnpackedTypes...>>
-{
-  typedef typename ApplyT::template apply<UnpackedTypes...> type;
-};
-
-template<typename ApplyT, typename... UnpackedTypes, typename... Types, typename... OtherTypeLists>
-struct CombineTypes<ApplyT, UnpackedTypeList<UnpackedTypes...>, ParameterList<Types...>, OtherTypeLists...>
-{
-  typedef CombineTypes<ApplyT, UnpackedTypeList<UnpackedTypes...>, ParameterList<Types...>, OtherTypeLists...> ThisT;
-
-  template<typename T1>
-  struct type_unpack
-  {
-    typedef UnpackedTypeList<UnpackedTypes..., T1> unpacked_t;
-    typedef CombineTypes<ApplyT, unpacked_t, OtherTypeLists...> combined_t;
-  };
-  
-  typedef ParameterList<typename ThisT::template type_unpack<Types>::combined_t::type...> type;
-};
-
+/// Primary specialization: Entry point that begins recursion on the first ParameterList.
+/// Unpacks the first list and creates recursive CombineTypes instantiations for each type,
+/// pairing it with all remaining lists.
+///
+/// Example instantiation:
+///   CombineTypes<ApplyType<Foo3>, ParameterList<int32_t, double>, ParameterList<P1,P2,bool>, ParameterList<float>>
+///   Types = {int32_t, double}
+///   OtherTypeLists = {ParameterList<P1,P2,bool>, ParameterList<float>}
+///   type = ParameterList<combined_t<int32_t>, combined_t<double>>
+///   where combined_t<int32_t> = CombineTypes<ApplyType<Foo3>, UnpackedTypeList<int32_t>, ParameterList<P1,P2,bool>, ParameterList<float>>
 template<typename ApplyT, typename... Types, typename... OtherTypeLists>
 struct CombineTypes<ApplyT, ParameterList<Types...>, OtherTypeLists...>
 {
-  typedef CombineTypes<ApplyT, ParameterList<Types...>, OtherTypeLists...> ThisT;
-
   template<typename T1>
-  struct type_unpack
-  {
-    typedef UnpackedTypeList<T1> unpacked_t;
-    typedef CombineTypes<ApplyT, unpacked_t, OtherTypeLists...> combined_t;
-  };
+  using combined_t = typename CombineTypes<ApplyT, UnpackedTypeList<T1>, OtherTypeLists...>::type;
 
-  typedef ParameterList<typename ThisT::template type_unpack<Types>::combined_t::type...> type;
+  using type = typename FlattenParameterLists<combined_t<Types>...>::flatten;
+};
+
+/// Secondary specialization: Accumulates types from remaining ParameterLists.
+/// Takes types already accumulated in UnpackedTypeList, unpacks the next ParameterList,
+/// and recursively combines each new type with the accumulated types.
+///
+/// Example instantiation:
+///   CombineTypes<ApplyType<Foo3>, UnpackedTypeList<int32_t>, ParameterList<P1,P2,bool>, ParameterList<float>>
+///   UnpackedTypes = {int32_t}
+///   Types = {P1, P2, bool}
+///   OtherTypeLists = {ParameterList<float>}
+///   type = ParameterList<combined_t<P1>, combined_t<P2>, combined_t<bool>>
+///   where combined_t<P1> = CombineTypes<ApplyType<Foo3>, UnpackedTypeList<int32_t, P1>, ParameterList<float>>
+template<typename ApplyT, typename... UnpackedTypes, typename... Types, typename... OtherTypeLists>
+struct CombineTypes<ApplyT, UnpackedTypeList<UnpackedTypes...>, ParameterList<Types...>, OtherTypeLists...>
+{
+  template<typename T1>
+  using combined_t = typename CombineTypes<ApplyT, UnpackedTypeList<UnpackedTypes..., T1>, OtherTypeLists...>::type;
+
+  using type = ParameterList<combined_t<Types>...>;
+};
+
+/// Base case: All ParameterLists exhausted, apply ApplyT to the accumulated type combination.
+/// UnpackedTypes contains one complete combination of types selected from each input ParameterList.
+///
+/// Example instantiation:
+///   CombineTypes<ApplyType<Foo3>, UnpackedTypeList<int32_t, P1, float>>
+///   UnpackedTypes = {int32_t, P1, float}
+///   type = Foo3<int32_t, P1, float>
+template<typename ApplyT, typename... UnpackedTypes>
+struct CombineTypes<ApplyT, UnpackedTypeList<UnpackedTypes...>>
+{
+  using type = typename ApplyT::template apply<UnpackedTypes...>;
 };
 
 // Default ApplyT implementation
@@ -1073,7 +1080,7 @@ struct DownCast
   {
     mod.method("cxxdowncast", [](SingletonType<DerivedT>, SuperT* base)
     {
-      if constexpr (std::is_polymorphic<DerivedT>::value)
+      if constexpr (std::is_polymorphic_v<DerivedT>)
       {
         return dynamic_cast<DerivedT*>(base);
       }
@@ -1083,7 +1090,7 @@ struct DownCast
       }
     });
     using newsuper_t = supertype<SuperT>;
-    if constexpr (!std::is_same<newsuper_t,SuperT>::value)
+    if constexpr (!std::is_same_v<newsuper_t,SuperT>)
     {
       DownCast<supertype<SuperT>, DerivedT>::apply(mod);
     }
@@ -1094,18 +1101,21 @@ template<typename T>
 inline void add_default_methods(Module& mod)
 {
   mod.set_override_module(get_cxxwrap_module());
-  if (!std::is_same<supertype<T>, T>::value)
+  if constexpr(!std::is_same_v<supertype<T>, T>)
   {
     mod.method("cxxupcast", UpCast<T>::apply);
     DownCast<supertype<T>,T>::apply(mod);
   }
-  if constexpr(std::is_destructible<T>::value)
+  if constexpr(std::is_destructible_v<T>)
   {
     mod.method("__delete", Finalizer<T>::finalize);
   }
   mod.unset_override_module();
 }
 
+template<typename T, typename... AppliedTypesT>
+class ParametricTypeWrappers;
+
 /// Helper class to wrap type methods
 template<typename T>
 class TypeWrapper
@@ -1168,7 +1178,7 @@ public:
 
   /// Define a "member" function using a lambda
   template<typename LambdaT, typename... Extra,
-           std::enable_if_t<detail::has_call_operator<LambdaT>::value && !std::is_member_function_pointer<LambdaT>::value, bool> = true>
+           std::enable_if_t<detail::has_call_operator<LambdaT>::value && !std::is_member_function_pointer_v<LambdaT>, bool> = true>
   TypeWrapper<T>& method(const std::string& name, LambdaT&& lambda, Extra... extra)
   {
     detail::ExtraFunctionData extraData = detail::parse_attributes(extra...);
@@ -1203,11 +1213,18 @@ public:
     return *this;
   }
 
+  template<typename... AppliedTypesT>
+  ParametricTypeWrappers<T, AppliedTypesT...> apply()
+  {
+    static_assert(detail::IsParametric<T>::value, "Apply can only be called on parametric types");
+    return ParametricTypeWrappers<T, AppliedTypesT...>(*this);
+  }
+
   template<typename... AppliedTypesT, typename FunctorT>
   TypeWrapper<T>& apply(FunctorT&& apply_ftor)
   {
     static_assert(detail::IsParametric<T>::value, "Apply can only be called on parametric types");
-    auto dummy = {this->template apply_internal<AppliedTypesT>(std::forward<FunctorT>(apply_ftor))...};
+    apply<AppliedTypesT...>().apply(std::forward<FunctorT>(apply_ftor));
     return *this;
   }
 
@@ -1231,42 +1248,11 @@ public:
 
 private:
 
-  template<typename AppliedT, typename FunctorT>
-  int apply_internal(FunctorT&& apply_ftor)
-  {
-    static constexpr int nb_julia_parameters = parameter_list<T>::nb_parameters;
-    static constexpr int nb_cpp_parameters = parameter_list<AppliedT>::nb_parameters;
-    static_assert(nb_cpp_parameters != 0, "No parameters found when applying type. Specialize jlcxx::BuildParameterList for your combination of type and non-type parameters.");
-    static_assert(nb_cpp_parameters >= nb_julia_parameters, "Parametric type applied to wrong number of parameters.");
-    const bool is_abstract = jl_is_abstracttype(m_dt);
-
-    detail::create_parameter_types<nb_julia_parameters>(parameter_list<AppliedT>(), std::make_index_sequence<nb_cpp_parameters>());
-
-    jl_datatype_t* app_dt = (jl_datatype_t*)apply_type((jl_value_t*)m_dt, parameter_list<AppliedT>()(nb_julia_parameters));
-    jl_datatype_t* app_box_dt = (jl_datatype_t*)apply_type((jl_value_t*)m_box_dt, parameter_list<AppliedT>()(nb_julia_parameters));
-
-    if(has_julia_type<AppliedT>())
-    {
-      std::cout << "existing type found : " << app_box_dt << " <-> " << julia_type<AppliedT>() << std::endl;
-      assert(julia_type<AppliedT>() == app_box_dt);
-    }
-    else
-    {
-      set_julia_type<AppliedT>(app_box_dt);
-      m_module.register_type(app_box_dt);
-    }
-    m_module.add_default_constructor<AppliedT>(app_dt);
-    m_module.add_copy_constructor<AppliedT>(app_dt);
-
-    apply_ftor(TypeWrapper<AppliedT>(m_module, app_dt, app_box_dt));
-
-    add_default_methods<AppliedT>(m_module);
-
-    return 0;
-  }
   Module& m_module;
   jl_datatype_t* m_dt;
   jl_datatype_t* m_box_dt;
+
+  template<typename U, typename... AppliedTypesT> friend class ParametricTypeWrappers;
 };
 
 using TypeWrapper1 = TypeWrapper<Parametric<TypeVar<1>>>;
@@ -1288,8 +1274,7 @@ template<typename T>
 template<typename ApplyT, typename... TypeLists, typename FunctorT>
 void TypeWrapper<T>::apply_combination(FunctorT&& ftor)
 {
-  typedef typename CombineTypes<ApplyT, TypeLists...>::type applied_list;
-  detail::DoApply<applied_list>()(*this, std::forward<FunctorT>(ftor));
+  detail::DoApply<combine_types<ApplyT, TypeLists...>>()(*this, std::forward<FunctorT>(ftor));
 }
 
 template<typename T, typename SuperParametersT, typename JLSuperT>
@@ -1297,7 +1282,7 @@ TypeWrapper<T> Module::add_type_internal(const std::string& name, JLSuperT* supe
 {
   static constexpr bool is_parametric = detail::IsParametric<T>::value;
   static_assert(!IsMirroredType<T>::value, "Mirrored types (marked with IsMirroredType) can't be added using add_type, map them directly to a struct instead and use map_type or explicitly disable mirroring for this type, e.g. define template<> struct IsMirroredType<Foo> : std::false_type { };");
-  static_assert(!std::is_scalar<T>::value, "Scalar types must be added using add_bits");
+  static_assert(!std::is_scalar_v<T>, "Scalar types must be added using add_bits");
 
   if(get_constant(name) != nullptr)
   {
@@ -1377,6 +1362,97 @@ TypeWrapper<T> Module::add_type(const std::string& name, JLSuperT* super)
   return add_type_internal<T, SuperParametersT>(name, super);
 }
 
+/// Collection of the wrappers of each class template instantiation, that is on
+/// Julia side, each fully-parametrized type.
+///
+/// Use the TypeWrapper::apply<...>() method to produce this collection.
+///
+/// The method ParametricTypeWrappers::apply(FunctorT&&) can then be used to add
+/// the wrappers for the class template methods.
+///
+template<typename T, typename... AppliedTypesT>
+class ParametricTypeWrappers
+{
+public:
+  ParametricTypeWrappers(TypeWrapper<T>& base_wrapper):
+    m_generic_type(base_wrapper),
+    m_specialized_types(std::make_tuple(create_type<AppliedTypesT>()...))
+  {
+  }
+
+  ///Apply a functor, ftor, to each wrapper.
+  ///
+  ///Usage example.
+  ///
+  ///  Wrap methods, `method1` and `method2` of a class template.
+  ///
+  ///  ```
+  ///  wrappers.apply([](auto wrapped)
+  ///  {
+  ///    typedef typename decltype(wrapped)::type WrappedT;
+  ///    wrapped.method("method1", &WrappedT::method1);
+  ///    wrapped.method("method2", &WrappedT::method2);
+  ///  });
+  ///  ```
+
+  template<typename FunctorT>
+  void apply(FunctorT&& ftor)
+  {
+    ssize_t i = 0;
+    apply_internal(std::forward<FunctorT>(ftor), m_specialized_types,
+                   std::index_sequence_for<AppliedTypesT...>{});
+  }
+
+private:
+
+  template<typename FunctorT, typename Tuple, std::size_t... I>
+  constexpr void apply_internal(FunctorT&& ftor, Tuple&& t, std::index_sequence<I...>)
+  {
+    //note: cast to decayed type required to allow in the functor statements like
+    //`using WrappedT = typename TypeWrapperT::type`
+    (void(ftor(static_cast<std::decay_t<decltype(std::get<I>(t))>>(std::get<I>(t)))), ...);
+  }
+
+  template<typename AppliedT>
+  TypeWrapper<AppliedT>
+  create_type()
+  {
+    static_assert(!IsMirroredType<AppliedT>::value, "Mirrored type templates can't be added using add_type and apply, you should explicitly disable mirroring for this type, e.g. define template<typename... T> struct IsMirroredType<Foo<T...>> : std::false_type { }; in the jlcxx namespace.");
+
+    static constexpr int nb_julia_parameters = parameter_list<T>::nb_parameters;
+    static constexpr int nb_cpp_parameters = parameter_list<AppliedT>::nb_parameters;
+    static_assert(nb_cpp_parameters != 0, "No parameters found when applying type. Specialize jlcxx::BuildParameterList for your combination of type and non-type parameters.");
+    static_assert(nb_cpp_parameters >= nb_julia_parameters, "Parametric type applied to wrong number of parameters.");
+    const bool is_abstract = jl_is_abstracttype(m_generic_type.dt());
+
+    detail::create_parameter_types<nb_julia_parameters>(parameter_list<AppliedT>(), std::make_index_sequence<nb_cpp_parameters>());
+
+    jl_datatype_t* app_dt = (jl_datatype_t*)apply_type((jl_value_t*)m_generic_type.dt(), parameter_list<AppliedT>()(nb_julia_parameters));
+    jl_datatype_t* app_box_dt = (jl_datatype_t*)apply_type((jl_value_t*)m_generic_type.m_box_dt, parameter_list<AppliedT>()(nb_julia_parameters));
+
+    if(has_julia_type<AppliedT>())
+      {
+        std::cout << "existing type found : " << app_box_dt << " <-> " << julia_type<AppliedT>() << std::endl;
+        assert(julia_type<AppliedT>() == app_box_dt);
+      }
+    else
+      {
+        set_julia_type<AppliedT>(app_box_dt);
+        m_generic_type.module().register_type(app_box_dt);
+      }
+
+    m_generic_type.module().template add_default_constructor<AppliedT>(app_dt);
+    m_generic_type.module().template add_copy_constructor<AppliedT>(app_dt);
+    add_default_methods<AppliedT>(m_generic_type.module());
+
+    return TypeWrapper<AppliedT>(m_generic_type.module(), app_dt, app_box_dt);
+  }
+
+private:
+  TypeWrapper<T>& m_generic_type;
+  std::tuple<TypeWrapper<AppliedTypesT>...> m_specialized_types;
+};
+
 namespace detail
 {
   template<typename T, bool>
@@ -1408,7 +1484,7 @@ void Module::add_bits(const std::string& name, JLSuperT* super)
 {
   assert(jl_is_datatype(super));
   static constexpr bool is_parametric = detail::IsParametric<T>::value;
-  static_assert(std::is_scalar<T>::value, "Bits types must be a scalar type");
+  static_assert(std::is_scalar_v<T>, "Bits types must be a scalar type");
   jl_svec_t* params = is_parametric ? parameter_list<T>()() : jl_emptysvec;
   JL_GC_PUSH1(&params);
   jl_datatype_t* dt = new_bitstype(jl_symbol(name.c_str()), m_jl_mod, (jl_datatype_t*)super, params, 8*sizeof(T));
